import multiprocessing


import pickle
from generator import Generator, Parameters, SurrogateDataset
from reconstruction import Reconstruction
from surrogate import Surrogate
from optimizer import Optimizer
from matplotlib import pyplot as plt
import time
import torch
import numpy as np
import os

def reset_weights(m):
    '''
    Try resetting model weights.
    '''
    for layer in m.children():
        if hasattr(layer, 'reset_parameters'):
            print(f'Reset trainable parameters of layer = {layer}')
            layer.reset_parameters()

if __name__ == "__main__":
    multiprocessing.set_start_method('spawn')

    outpath = 'run2_EM_homo_simple3'
    
    os.system('mkdir  '+outpath)
    outpath += '/'
    os.system('cp *.py '+outpath)

    test = False


    divide = 10 if test else 1

    #so far these are only layer thicknesses
    start_pars = {'thickness_absorber_0': .1,
                  #'thickness_absorber_1': .1, 
                  #'thickness_absorber_2': .1, 
                  #'thickness_absorber_3': .1, 
                  #'thickness_absorber_4': .1, 
                  #'thickness_absorber_5': .1, 
                  #'thickness_absorber_6': .1, 
                  #'thickness_absorber_7': .1, 
                  #'thickness_absorber_8': .1, 
                  'thickness_scintillator_0': 0.5, 
                  #'thickness_scintillator_1': 0.5,
                  #'thickness_scintillator_2': 0.5,
                  #'thickness_scintillator_3': 0.5,
                  #'thickness_scintillator_4': 0.5,
                  #'thickness_scintillator_5': 0.5,
                  #'thickness_scintillator_6': 0.5,
                  #'thickness_scintillator_7': 0.5,
                  #'thickness_scintillator_8': 0.5,
                  }
    box = np.array(len(start_pars)*[1.5]) #will be adjusted to some extent by optimizer
    #start parameters from run 8
    #start_pars = {'thickness_absorber_0': 0.7642903, 'thickness_absorber_1': 10.469371, 'thickness_scintillator_0': 30.585306, 'thickness_scintillator_1': 22.256506}
    
    
    gen = Generator(box, 
                    Parameters(parameters = start_pars),
                    n_vars = 30,
                    n_events_per_var = 400//divide,
                    particles=[['gamma',0.22]])#,['pi+',2.11]])#0)
    
    
    # reconstruction
    reco_model = Reconstruction(gen.n_parameters, 
                           gen.n_parameters_per_sensor * gen.n_sensors, 
                           gen.n_target_parameters)
    
    surrogate_model = Surrogate(gen.n_parameters, 
                           gen.n_target_parameters + gen.n_context_parameters, #context
                           gen.n_target_parameters,  #reco
                           n_time_steps=100)
    
    optimizer = Optimizer(gen, surrogate_model, reco_model, gen.parameters,
                          constraints= {'length': 25}) #one meter
    
    n_epochs_pre = 30//divide
    n_epochs_main = 100//divide
    
    parameters = gen.parameters

    #if saved evolution exists, load it
    import os.path
    if False and os.path.isfile('evolution.npy'):
        evolution = np.load('evolution.npy').tolist()
        print('EVOLUTION:',evolution)
        parameters = evolution[-1]
    
    evolution = []
    def save_evolution(parameters, losses, reco_losses, reco_model, means, std):
        '''
        add to evolution list and save it with pickle.
        save it in human readable format as well
        '''
        ps_dict = gen.translate_parameters(parameters)
        ps = {k: ps_dict[k] for k in ps_dict.keys()}
        ps.update(
            {'reco_model': reco_model.state_dict(),
             'means': means,
             'stds': std}
        )
        evolution.append([ps, losses, reco_losses])
        with open(outpath+'evolution.pkl', 'wb') as f:
            pickle.dump(evolution, f)
        # human readable
        with open(outpath+'evolution.txt', 'w') as f:
            for e in evolution:
                f.write(str(e[0]) + ' ' + str(e[1]) + ' ' + str(e[2]) + '\n')

        # now plot all entries of the dictionary ps in a plot and save it, make a new point for each of the items in the evolution list and use the dictionary keys as legend
        # 'transpose' dictionary
        pltdict = {k: [] for k in ps_dict.keys()}
        for i in range(len(evolution)):
            for k in pltdict.keys():
                pltdict[k] += [evolution[i][0][k]]

        # now make a plot with matplotlib
        for k in pltdict.keys():
            style = '-'
            if 'absorber' in k:
                style = '--'
            plt.plot([i for i in range(len(pltdict[k]))], np.where(np.array(pltdict[k]) < 0., 0., np.array(pltdict[k])), label = k, linestyle=style)
        
        plt.legend()
        # create a second y-axis for the loss values. Make the line black and thicker than the others; make the loss y axis log scale
        plt.twinx()
        plt.plot([i for i in range(len(evolution))], [e[1] for e in evolution], label = 'loss', color='black', linewidth=2, alpha=0.2)
        #also plot reco loss also in black but with a different linestyle
        plt.plot([i for i in range(len(evolution))], [e[2] for e in evolution], label = 'reco loss', color='black', linewidth=2, alpha=0.2, linestyle='--')
        plt.yscale('log')
        
        plt.legend()
        plt.xlabel('step')
        plt.savefig(outpath+'evolution.png')
        plt.close()




    #save_evolution(parameters, 1., 1.)
    #exit()
    
    def make_check_plot(surr_out, reco_out, true_in, evolution, outname="" ):
        '''
        make response plots for all three and save them, use evolution size as step index for output file name
        '''
        reco_resp = (reco_out.detach().cpu().numpy() - true_in.detach().cpu().numpy()) / true_in.detach().cpu().numpy()
        surr_resp = (surr_out.detach().cpu().numpy() - true_in.detach().cpu().numpy()) / true_in.detach().cpu().numpy()
    
        #range = np.min([np.min(reco_resp), np.min(surr_resp)]), np.max([np.max(reco_resp), np.max(surr_resp)])

        if len(outname) == 0:
            outname = str(len(evolution))
    
        plt.figure(figsize=(10,10))
        plt.hist(reco_resp.flatten(), bins=51, range=(-1,1), alpha = 0.5, label='G4+reco')
        plt.hist(surr_resp.flatten(), bins=51, range=(-1,1), alpha = 0.5, label='E2E surrogate')
        plt.text(0.15, 0.8, 'step '+outname, transform=plt.gca().transAxes)
        if len(evolution):
            plt.text(0.15, 0.7, 'o_loss '+str((evolution[-1][1])), transform=plt.gca().transAxes)
        plt.legend()
        plt.savefig(outpath+'response_'+outname+'_lin.png')
        plt.yscale('log')
        plt.savefig(outpath+'response_'+outname+'_log.png')
        plt.close()

    #pre-train reco and surrogate model
    def pre_train():
        ds = gen.create_torch_dataset(parameters, random_scaler=1.2)
        reco_model.to('cuda')
        print('pre-training 0')
        reco_model.train_model(ds, batch_size=256, n_epochs= 10, lr=0.03)
        print('pre-training 1')
        reco_model.train_model(ds, batch_size=256, n_epochs= n_epochs_pre, lr=0.01)
        print('pre-training 2')
        reco_model.train_model(ds, batch_size=512, n_epochs= n_epochs_pre, lr=0.001)
        print('pre-training 3')
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_pre, lr=0.001)
        #this is a normalised quantity - at some point only expose the un-normalised ones, but also needs adjustments for surrogate dataset
        reco_result,reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)

        reco_model.to('cpu')
    
        surrogate_dataset = SurrogateDataset(ds, reco_result.detach().cpu().numpy())#important to detach here
        print('surr pre-training 0')
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= 10, lr=0.03)
        print('surr pre-training 1')
        surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_pre, lr=0.01)
        print('surr pre-training 2')
        surrogate_model.train_model(surrogate_dataset, batch_size=512, n_epochs= n_epochs_pre, lr=0.001)
        print('surr pre-training 3')
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_pre, lr=0.001)
        print('surr pre-training 4')
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_pre, lr=0.0003)
        #these are un-normalised quantities
        surr_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=512)
        make_check_plot(surr_out, reco_out, true_in, evolution, outname = "pretrain")
    pre_train()
    

    best_surrogate_loss = 1e10
    def update_best_surrogate_loss(loss):
        global best_surrogate_loss
        if loss < best_surrogate_loss:
            best_surrogate_loss = loss
            return True
        return loss < 4. * best_surrogate_loss

    covariance_pcas = []
    
    for i in range(1000): #just an example
        
        ds = gen.create_torch_dataset(parameters) # full variations plus a tiny bit
        reco_model.to('cuda')
        reco_model.train_model(ds, batch_size=256, n_epochs= n_epochs_main//4, lr=0.003)
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.001)
        reco_model.train_model(ds, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0003)
        reco_result, reco_loss = reco_model.apply_model_in_batches(ds, batch_size=128)
        reco_model.to('cpu') #free up gpu memory
    
        #create the surrogate dataset
        surrogate_dataset = SurrogateDataset(ds, reco_result.detach().cpu().numpy())#important to detach here
        
        print("Surrogate training")
        surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.005) #get the surrogate model to a good state
        sl = surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main, lr=0.0003) #get the surrogate model to a good state
    
        while not update_best_surrogate_loss(sl):
            print("Surrogate retraining")
            pre_train()
            surrogate_model.train_model(surrogate_dataset, batch_size=256, n_epochs= n_epochs_main//5, lr=0.005)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.005)
            surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0003)
            sl = surrogate_model.train_model(surrogate_dataset, batch_size=1024, n_epochs= n_epochs_main//2, lr=0.0001)

        ## check if things worked
        surr_out, reco_out, true_in = surrogate_model.apply_model_in_batches(surrogate_dataset, batch_size=512)
        make_check_plot(surr_out, reco_out, true_in, evolution)
    
        print("Optimization")
        #apply the surrogate model
        updated_detector_parameters, optimal, o_loss = optimizer.optimize(surrogate_dataset, batch_size=512, n_epochs=40, lr=0.02, add_constraints=True)#some noise
        print("Updated detector parameters: ", gen.translate_parameters(updated_detector_parameters), "; not interrupted", optimal, "; change", updated_detector_parameters - parameters)
        

        #use the difference of old and updated parameters to define a direction in parameter space
        #and increase the box size in that direction
        #gen.box_size += updated_detector_parameters - parameters
        # make sure the size does not become too small in any dimension
        #gen.box_size = np.maximum(gen.box_size, 0.1 * np.ones_like(gen.box_size))

        if np.isnan(o_loss) or np.isnan(updated_detector_parameters).any() or np.isinf(o_loss) or np.isinf(updated_detector_parameters).any():
            print("NaN loss, restarting.")
            surrogate_model.apply(reset_weights)
            reco_model.apply(reset_weights)
            pre_train()
            sl = 0. #reset surrogate loss too to avoid infinite loop
            continue
        #
        # step more in the direction. 
        parameters = updated_detector_parameters + 0.5*(updated_detector_parameters - parameters)


        # for bookkeeping, append the generator box_covariance to a text file
        with open(outpath+'box_covariance.txt', 'a') as f:
            f.write(str(gen.box_covariance) + '\n')
        
        # also perform a quick PCA of the box covariance and plot the first two principal components
        # but for all entries of covariance_pcas
        pca = np.linalg.eig(gen.box_covariance)
        covariance_pcas.append(pca)
        plt.plot(range(len(covariance_pcas)), [p[0][0] for p in covariance_pcas])
        plt.plot(range(len(covariance_pcas)), [p[0][1] for p in covariance_pcas])
        plt.xlabel('Step')
        plt.ylabel('1st & 2nd principal component')
        plt.savefig(outpath+f'pca.png')#just overwrite
        plt.close()


        if test:
            print('EVOLUTION:',evolution, gen.box_covariance)
            time.sleep(2)
        save_evolution(updated_detector_parameters, o_loss, reco_loss, reco_model, gen.means, gen.stds) #save the optimised parameters here, not the ones for the next box generation
        #exit()

    
    

    
